from __future__ import annotations
from collections import defaultdict
import re
from dataclasses import dataclass, asdict, field
from hashlib import md5
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Tuple, Union, Set
from urllib.parse import urlparse

import fastnumbers
import orjson
from slugify import slugify

import sm.misc as M
from sm.inputs.table import ColumnBasedTable, Column
from grams.inputs.context import ContentHierarchy


@dataclass
class LinkedTable:
    # the table that we are working on
    table: ColumnBasedTable

    # context of the table
    context: Context

    # a mapping from (row id, column id) to the list of links attach to that cell (not include header)
    links: List[List[List[Link]]]

    @property
    def id(self):
        return self.table.table_id

    def shape(self):
        return self.table.shape()

    def remove_nonexistent_entities(self, nonexisting_entities: Set[str]):
        nrows, ncols = self.table.shape()
        for ri in range(nrows):
            for ci in range(ncols):
                for link in self.links[ri][ci]:
                    link.remove_nonexisting_entities(nonexisting_entities)

    def size(self):
        """Number of rows of the table"""
        if len(self.table.columns) == 0:
            return 0
        return len(self.table.columns[0].values)

    def iter_cell_index(self):
        nrows, ncols = self.table.shape()
        for ri in range(nrows):
            for ci in range(ncols):
                yield ri, ci

    def to_dict(self):
        return {
            "version": "1.2",
            "table": self.table.to_dict(),
            "context": self.context.to_dict(),
            "links": [
                [[link.to_dict() for link in links] for links in rlinks]
                for rlinks in self.links
            ],
        }

    def get_friendly_fs_id(self):
        return LinkedTable._get_friendly_fs_id(self.id)

    @staticmethod
    def _get_friendly_fs_id(id: str):
        if id.startswith("http://") or id.startswith("https://"):
            if id.find("dbpedia.org") != -1:
                return (
                    slugify(
                        urlparse(id).path.replace("/resource/", "").replace("/", "_")
                    ).replace("-", "_")
                    + "_"
                    + md5(id.encode()).hexdigest()
                )

            if id.find("wikipedia.org") != -1:
                return (
                    slugify(
                        urlparse(id).path.replace("/wiki/", "").replace("/", "_")
                    ).replace("-", "_")
                    + "_"
                    + md5(id.encode()).hexdigest()
                )

            raise NotImplementedError()
        return slugify(id.replace("/", "_")).replace("-", "_")

    @staticmethod
    def from_dict(odict: dict):
        assert float(odict.get("version", "0.0")) >= 1.1
        tbl = ColumnBasedTable.from_dict(odict["table"])
        context = Context.from_dict(odict["context"])
        links = [
            [[Link.from_dict(link) for link in links] for links in rlinks]
            for rlinks in odict["links"]
        ]
        return LinkedTable(tbl, context, links)

    @staticmethod
    def from_column_based_table(tbl: ColumnBasedTable):
        links = []
        if len(tbl.columns) > 0:
            links = [
                [[] for ci in range(len(tbl.columns))]
                for ri in range(len(tbl.columns[0].values))
            ]
        return LinkedTable(tbl, Context(None, None, None), links)

    @staticmethod
    def from_csv_file(
        infile: Union[Path, str],
        link_file: Optional[Union[Path, str]] = None,
        first_row_header: bool = True,
        table_id: Optional[str] = None,
        top_k: int = 100,
    ) -> LinkedTable:
        """Load table from a csv file, and its links from a tsv file (if exist).

        For format of the link file, see `LinkedTable.parse_link_file`

        Args:
            infile: csv file
            link_file: if not provided, this function will look for a file of same name of  `infile` in the same folder but ends with `.links.tsv`.
            first_row_header: whether the first row of the `infile` is header of the table
            table_id: if None, the table id will be file name of `infile`
            top_k: top k candidate entity to keep.
        """
        infile = Path(infile)
        if link_file is None:
            link_file = infile.parent / f"{infile.stem}.links.tsv"
        else:
            link_file = Path(link_file)

        if table_id is None:
            table_id = infile.stem
        rows = M.deserialize_csv(infile)

        assert len(rows) > 0, "Empty table"
        columns = []
        if first_row_header:
            headers = rows[0]
            rows = rows[1:]
        else:
            headers = [f"column-{i:03d}" for i in range(len(rows[0]))]

        for ci, cname in enumerate(headers):
            columns.append(Column(ci, cname, [r[ci] for r in rows]))
        table = ColumnBasedTable(table_id, columns)

        if link_file.exists():
            links = LinkedTable.parse_link_file(table, link_file, top_k)
            for rlinks in links:
                for clinks in rlinks:
                    for link in clinks:
                        if link.entity_id is not None:
                            link.candidates = [CandidateEntity(link.entity_id, 1.0)]
        else:
            links = []
            for ri in range(len(rows)):
                links.append([[] for _ci in range(len(headers))])

        return LinkedTable(table, Context(), links)

    @staticmethod
    def parse_link_file(
        table: ColumnBasedTable, infile: Union[Path, str], top_k: int = 100
    ) -> List[List[List[Link]]]:
        """
        Each row of a link file has the following format: `<row_index>\t<col_index>\t(<link>|(<entity_id>(\t<entity_id>)*))`, where:
            * `row_index` and `col_index` start from 0
            * `row_index` does not count the header of the table (i.e., skip the first row of `infile` if it's the header)
            * `(<link>|(<entity_id>(\t<entity_id>)*))` is either:
                - `<link>` a json string encoding Link object and can be deserialized using `Link.from_dict` function
                - or (<entity_id>(\t<entity_id>)*) a list of entity ids joined by `\t` tab character, each entity id can be a wikidata qnode id (e.g., Q414) or a full qnode uri (e.g., "http://www.wikidata.org/entity/Q414"). The first entity is considered as the correct entity of the cell, and the rest are considered as the candidate entities of the cell
        Note that a pair `<row_index>`, `<col_index>` don't have to be unique.
        """
        links = []
        nrows, ncols = table.shape()

        for ri in range(nrows):
            links.append([[] for _ci in range(ncols)])

        rows = []
        for row in M.deserialize_csv(infile, delimiter="\t"):
            ri, ci, ents = int(row[0]), int(row[1]), row[2:]
            rows.append((ri, ci, ents))

        has_only_correct_entity = all(len(ents) == 1 for _, _, ents in rows)

        for ri, ci, ents in rows:
            if len(ents) == 1 and ents[0].startswith("{"):
                link = Link.from_dict(orjson.loads(ents[0]))
            else:
                pents = []
                for ent in ents[: top_k + 1]:
                    if ent.startswith("http"):
                        assert ent.startswith("http://www.wikidata.org/entity/")
                        qnode_id = ent.replace("http://www.wikidata.org/entity/", "")
                    else:
                        qnode_id = ent

                    if ":" in qnode_id:
                        qnode_id, qnode_prob = qnode_id.split(":", 1)
                        assert fastnumbers.isfloat(qnode_prob)
                        qnode_prob = fastnumbers.float(qnode_prob)
                    else:
                        qnode_prob = 1.0
                    pents.append((qnode_id, qnode_prob))

                if has_only_correct_entity:
                    candidates = [CandidateEntity(pents[0][0], pents[0][1])]
                else:
                    candidates = [CandidateEntity(e[0], e[1]) for e in pents[1:]]

                link = Link(
                    start=0,
                    end=len(table.columns[ci][ri]),
                    url=f"http://www.wikidata.org/entity/{pents[0][0]}",
                    entity_id=pents[0][0] if pents[0][0].strip() != "" else None,
                    candidates=candidates,
                )
            links[ri][ci].append(link)

        return links


@dataclass
class Context:
    """Table's context"""

    page_title: Optional[str] = None
    page_url: Optional[str] = None
    page_entity_id: Optional[str] = None

    content_hierarchy: List[ContentHierarchy] = field(default_factory=list)

    def to_dict(self):
        return {
            "page_title": self.page_title,
            "page_url": self.page_url,
            "page_entity_id": self.page_entity_id,
            "content_hierarchy": [c.to_dict() for c in self.content_hierarchy],
        }

    @staticmethod
    def from_dict(odict: dict):
        return Context(
            page_title=odict.get("page_title"),
            page_url=odict.get("page_url"),
            page_entity_id=odict.get("page_entity_id"),
            content_hierarchy=[
                ContentHierarchy.from_dict(c)
                for c in odict.get("content_hierarchy", [])
            ],
        )


@dataclass
class CandidateEntity:
    entity_id: str
    probability: float


@dataclass
class Link:
    start: int
    end: int  # exclusive
    url: Optional[str]  # none when there is no link, and the entity is not mapped yet
    entity_id: Optional[str]
    candidates: List[CandidateEntity]

    @staticmethod
    def from_dict(obj: dict) -> Link:
        return Link(
            start=obj["start"],
            end=obj["end"],
            url=obj["url"],
            entity_id=obj["entity_id"],
            candidates=[CandidateEntity(**c) for c in obj.get("candidates", [])],
        )

    def to_dict(self) -> dict:
        return {
            "start": self.start,
            "end": self.end,
            "url": self.url,
            "entity_id": self.entity_id,
            "candidates": [asdict(c) for c in self.candidates],
        }

    def remove_nonexisting_entities(self, nonexisting_entities: Set[str]):
        if self.entity_id in nonexisting_entities:
            self.entity_id = None
        self.candidates = [
            c for c in self.candidates if c.entity_id not in nonexisting_entities
        ]

    def length(self) -> int:
        return self.end - self.start



class DGConfigs:
    """Holding configuration for data graph construction"""

    # allow kg search for the same entity - whether we try to discover relationships between the same entity in same row but different columns
    # recommend to disable.
    ALLOW_SAME_ENT_SEARCH = False
    # removing unnecessary entities to keep the graph size reasonable
    PRUNE_REDUNDANT_ENT = True
    # remove leaf entities that is property of other nodes but the node doesn't have any other qualifiers
    PRUNE_SINGLE_LEAF_ENT = False
    # path discovering using index.
    # recommend to turn this on
    USE_KG_INDEX = True
    # enable context usage
    USE_CONTEXT = True



from typing import Tuple
from grams.algorithm.literal_matchers.string_match import string_match_similarity
from grams.algorithm.literal_matchers.text_parser import ParsedTextRepr
from kgdata.wikidata.models import WDValue
from grams.algorithm.literal_matchers.types import LiteralMatchKit


def entity_similarity_test(
    kgval: WDValue, val: ParsedTextRepr, kit: LiteralMatchKit
) -> Tuple[bool, float]:
    """Compare if the value in KG matches with value in the cell

    Args:
        kgval: the value in KG
        val: the value in the cell
        kit: objects that may be needed for matching
    Returns:
        a tuple of (whether it's matched, confidence)
    """

    if not kgval.is_qnode(kgval):
        # not handle lexical or property yet
        return False, 0.0

    qnode = kit.wdentities[kgval.as_entity_id()]
    lst = [x.strip() for x in val.normed_string.split(",")]
    for label in [str(qnode.label)] + [str(x) for x in qnode.aliases]:
        for item in lst:
            result = string_match_similarity(label, item)
            if result[0] is True:
                return result
    return False, 0.0



import warnings
from enum import IntEnum
from typing import Dict, Mapping

from kgdata.wikidata.models import WDEntity, WDClass, WDProperty
from sm.misc import identity_func

warnings.warn(
    "WDOnt is deprecated and will be removed in a future release.",
    DeprecationWarning,
)


class OutOfNamespace(Exception):
    pass


class WDOnt:
    STATEMENT_URI = "http://wikiba.se/ontology#Statement"
    STATEMENT_REL_URI = "wikibase:Statement"

    def __init__(
        self,
        qnodes: Mapping[str, WDEntity],
        qnode_labels: Mapping[str, str],
        wdclasses: Mapping[str, WDClass],
        wdprops: Mapping[str, WDProperty],
    ):
        self.qnodes = qnodes
        self.qnode_labels = qnode_labels
        self.wdclasses = wdclasses
        self.wdprops = wdprops
        self.get_qid_fn = {
            "Q": identity_func,
            "q": identity_func,
            # http
            "h": self.get_qnode_id,
        }
        self.get_pid_fn = {
            "P": identity_func,
            "p": identity_func,
            "h": self.get_prop_id,
        }

    @classmethod
    def is_uri_statement(cls, uri: str):
        return uri == "http://wikiba.se/ontology#Statement"

    @classmethod
    def is_uri_dummy_class(cls, uri: str):
        return uri == "http://wikiba.se/ontology#DummyClassForInversion"

    @classmethod
    def is_uri_property(cls, uri: str):
        return uri.startswith(f"http://www.wikidata.org/prop/")

    @classmethod
    def is_uri_qnode(cls, uri: str):
        return uri.startswith("http://www.wikidata.org/entity/")

    @classmethod
    def get_qnode_id(cls, uri: str):
        if not cls.is_uri_qnode(uri):
            raise OutOfNamespace(f"{uri} is not in wikidata qnode namespace")
        return uri.replace("http://www.wikidata.org/entity/", "")

    @classmethod
    def get_qnode_uri(cls, qnode_id: str):
        return f"http://www.wikidata.org/entity/{qnode_id}"

    @classmethod
    def get_qnode_rel_uri(cls, qnode_id: str):
        return f"wd:{qnode_id}"

    @classmethod
    def get_prop_id(cls, uri: str):
        if not cls.is_uri_property(uri):
            raise OutOfNamespace(f"{uri} is not in wikidata property namespace")
        return uri.replace(f"http://www.wikidata.org/prop/", "")

    @classmethod
    def get_prop_uri(cls, pid: str):
        return f"http://www.wikidata.org/prop/{pid}"

    @classmethod
    def get_prop_rel_uri(cls, pid: str):
        return f"p:{pid}"

    def get_qnode_label(self, uri_or_id: str):
        qid = self.get_qid_fn[uri_or_id[0]](uri_or_id)
        if qid in self.wdclasses:
            label = self.wdclasses[qid].label
        elif qid in self.qnodes:
            label = self.qnodes[qid].label
        else:
            label = self.qnode_labels.get(qid, qid)
        return f"{label} ({qid})"

    def get_pnode_label(self, uri_or_id: str):
        pid = self.get_pid_fn[uri_or_id[0]](uri_or_id)
        # TODO: fix me! should not do this
        if pid not in self.wdprops:
            return pid
        return f"{self.wdprops[pid].label} ({pid})"



import enum
import time
from collections import defaultdict
from pathlib import Path
from typing import Dict, Iterable, Mapping, Set, Union, List, NamedTuple, Optional

import sm.misc as M
from loguru import logger
from tqdm.auto import tqdm
from kgdata.wikidata.models import WDEntity, WDProperty

Relationship = NamedTuple(
    "Relationship", [("prop", str), ("quals", List[str]), ("both", bool)]
)
OneHopIndexPath = NamedTuple(
    "OneHopIndexPath", [("relationship", Relationship), ("statement_index", int)]
)
OneHopIndex = Dict[str, Dict[str, List[OneHopIndexPath]]]

# to find all paths between hop1 & hop2, you need to do cross product, but if you consider only transitive props, then
# you need to filter the second hop to be only equal
TwoHopIndexPath = NamedTuple(
    "TwoHopIndexPath",
    [("qnode", str), ("hop1", List[OneHopIndexPath]), ("hop2", List[OneHopIndexPath])],
)
TwoHopIndex = Dict[str, Dict[str, List[TwoHopIndexPath]]]


class TraversalOption(enum.Enum):
    NoConstraint = "no_constraint"
    TransitiveOnly = "transitive_only"


class KGObjectIndex:
    """An index in KG that help speed up searching for entities"""

    def __init__(
        self,
        one_hop_index: OneHopIndex,
        two_hop_index: TwoHopIndex,
        n_hop: int,
        traversal_option: TraversalOption,
    ):
        self.one_hop_index = one_hop_index
        self.two_hop_index = two_hop_index
        self.n_hop = n_hop
        self.traversal_option = traversal_option

    @staticmethod
    def from_entities(
        index_qnode_ids: List[str],
        entities: Mapping[str, WDEntity],
        wdprops: Mapping[str, WDProperty],
        n_hop: int = 1,
        traversal_option: TraversalOption = TraversalOption.TransitiveOnly,
        verbose: bool = False,
    ):
        # one hop index
        hop1_index: OneHopIndex = {}
        hop2_index: TwoHopIndex = {}

        assert 1 <= n_hop <= 2
        if n_hop > 1 and traversal_option == TraversalOption.TransitiveOnly:
            transitive_props = {pid for pid, p in wdprops.items() if p.is_transitive()}
        else:
            transitive_props = None

        pbar = tqdm(
            total=len(index_qnode_ids) * (1 + int(n_hop > 1)),
            desc=f"build kg object index for {len(index_qnode_ids)} qnodes",
            disable=not verbose,
        )

        for index_qnode_id in index_qnode_ids:
            qnode = entities[index_qnode_id]
            hop1_index[qnode.id] = dict(build_one_hop_index(qnode))
            pbar.update(1)

        if n_hop > 1:
            for index_qnode_id in index_qnode_ids:
                qnode_hop2_index = defaultdict(list)
                for uid, hop1_paths in hop1_index[index_qnode_id].items():
                    # TODO: ignore P31 and not discovering outside of it until to fix the method that discovers incoming&outgoing edges to include classes
                    # which currently has scalability issue since they have too many edges.
                    hop1_paths = [
                        hop1_path
                        for hop1_path in hop1_paths
                        if hop1_path.relationship.prop != "P31"
                        and (
                            transitive_props is None
                            or hop1_path.relationship.prop in transitive_props
                        )
                    ]
                    if len(hop1_paths) == 0:
                        continue
                    if transitive_props is not None:
                        hop1_props = {path.relationship.prop for path in hop1_paths}

                    if uid not in entities:
                        # this can happen due to some of the qnodes is in the link, but is missing in the KG
                        # this is very rare so we can employ some check to make sure this is not due to
                        # our wikidata subset
                        is_error_in_kg = any(
                            any(
                                _s.value.is_entity_id(_s.value)
                                and _s.value.as_entity_id() in entities
                                for _s in _stmts
                            )
                            for _p, _stmts in entities[index_qnode_id].props.items()
                        )
                        if not is_error_in_kg:
                            raise Exception(f"Missing qnodes in your KG subset: {uid}")
                        continue

                    for vid, hop2_paths in build_one_hop_index(
                        entities[uid], transitive_props
                    ).items():
                        hop2_paths = [
                            hop2_path
                            for hop2_path in hop2_paths
                            if transitive_props is None
                            or hop2_path.relationship.prop in hop1_props
                        ]
                        if len(hop2_paths) > 0:
                            qnode_hop2_index[vid].append(
                                TwoHopIndexPath(uid, hop1_paths, hop2_paths)
                            )
                pbar.update(1)
                hop2_index[index_qnode_id] = dict(qnode_hop2_index)
        pbar.close()
        return KGObjectIndex(hop1_index, hop2_index, n_hop, traversal_option)

    @staticmethod
    def deserialize(infile: Union[Path, str], verbose: bool = False):
        start = time.time()
        index = M.deserialize_pkl(infile)
        if verbose:
            logger.info(
                "Deserialize KG object index takes {} seconds",
                f"{time.time() - start:.3f}",
            )
        return index

    def serialize(self, outfile: Union[Path, str]):
        Path(outfile).parent.mkdir(exist_ok=True, parents=True)
        M.serialize_pkl(self, outfile)

    def iter_hop1_props(
        self, source_qnode_id: str, target_qnode_id: str
    ) -> Iterable[OneHopIndexPath]:
        return self.one_hop_index[source_qnode_id].get(target_qnode_id, [])

    def iter_hop2_props(
        self, source_qnode_id: str, target_qnode_id: str
    ) -> Iterable[TwoHopIndexPath]:
        return self.two_hop_index[source_qnode_id].get(target_qnode_id, [])


def build_one_hop_index(qnode: WDEntity, filter_props: Optional[Set[str]] = None):
    hop1_index = defaultdict(list)
    if filter_props is None:
        iter = qnode.props.items()
    else:
        iter = [(p, qnode.props.get(p, [])) for p in filter_props]

    for p, stmts in iter:
        for stmt_i, stmt in enumerate(stmts):
            lst = defaultdict(list)
            if stmt.value.is_entity_id(stmt.value):
                target_qnode_id = stmt.value.as_entity_id()
                lst[target_qnode_id].append(None)

            for q, qvals in stmt.qualifiers.items():
                for target_qnode_id in {
                    qval.as_entity_id() for qval in qvals if qval.is_entity_id(qval)
                }:
                    lst[target_qnode_id].append(q)

            for target_qnode_id, rels in lst.items():
                assert all(x is not None for x in rels[1:])
                # (p & q) is guarantee to be unique
                if rels[0] is not None or len(rels) > 1:
                    # we have qualifiers
                    qs = rels if rels[0] is not None else rels[1:]
                else:
                    qs = []
                rel = Relationship(p, qs, both=rels[0] is None and len(rels) > 1)
                hop1_index[target_qnode_id].append(OneHopIndexPath(rel, stmt_i))
    return hop1_index



import grams.inputs as I
from grams.main import GRAMS
from grams.html_table_parser import fetch_tables
from grams.config import ROOT_DIR, DATA_DIR
from grams.algorithm.sm_wikidata import WikidataSemanticModelHelper



from rltk.similarity import hybrid_jaccard_similarity


class StringSimilarity:
    @staticmethod
    def hybrid_jaccard_similarity(s1: str, s2: str) -> float:
        return hybrid_jaccard_similarity(set(s1.split(" ")), set(s2.split(" ")))



import re
from typing import Tuple, cast
from grams.algorithm.literal_matchers.text_parser import (
    ParsedTextRepr,
    ParsedDatetimeRepr,
)
from kgdata.wikidata.models import WDValueQuantity
from grams.algorithm.literal_matchers.types import LiteralMatchKit


def quantity_test(
    kgval: WDValueQuantity, val: ParsedTextRepr, kit: LiteralMatchKit
) -> Tuple[bool, float]:
    """Compare if the value in KG matches with value in the cell

    Args:
        kgval: the value in KG
        val: the value in the cell
        kit: objects that may be needed for matching

    Returns:
        a tuple of (whether it's matched, confidence)
    """
    if val.number is None:
        return False, 0.0

    kgnum = float(kgval.value["amount"])
    num = val.number.number

    if abs(kgnum - num) < 1e-5:
        return True, 1.0
    if kgnum == 0:
        diff_percentage = abs(kgnum - num) / 1e-7
    else:
        diff_percentage = abs((kgnum - num) / kgnum)

    if diff_percentage < 0.05:
        # within 5%
        return True, 0.95 - diff_percentage
    # if diff_percentage < 0.1:
    #     return True, 0.9 - diff_percentage
    return False, 0.0



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['grams',
 'grams.algorithm',
 'grams.algorithm.candidate_graph',
 'grams.algorithm.data_graph',
 'grams.algorithm.deprecated',
 'grams.algorithm.inferences',
 'grams.algorithm.inferences.features',
 'grams.algorithm.literal_matchers',
 'grams.algorithm.postprocessing',
 'grams.html_table_parser',
 'grams.inputs']

package_data = \
{'': ['*']}

install_requires = \
['beautifulsoup4>=4.9.3,<5.0.0',
 'click>=8.0.0,<=8.0.4',
 'fastnumbers>=3.2.1,<4.0.0',
 'ftfy>=6.1.1,<7.0.0',
 'graph-wrapper>=1.5.0,<2.0.0',
 'html5lib>=1.1,<2.0',
 'hugedict>=2.7.1,<3.0.0',
 'ipython>=8.0.1,<9.0.0',
 'kgdata>=3.2.0,<4.0.0',
 'loguru>=0.6.0,<0.7.0',
 'matplotlib>=3.5.3,<4.0.0',
 'ned>=1.0.0,<2.0.0',
 'networkx>=2.8.2,<3.0.0',
 'numparser2>=1.0.0,<2.0.0',
 'omegaconf>=2.0.6,<3.0.0',
 'orjson>=3.8.0,<4.0.0',
 'pslpython==2.2.2',
 'python-slugify>=6.1.2,<7.0.0',
 'rdflib>=6.1.1,<7.0.0',
 'requests>=2.28.0,<3.0.0',
 'rltk==2.0.0a20',
 'ruamel.yaml>=0.17.21,<0.18.0',
 'sem-desc>=3.9.0,<4.0.0',
 'steiner-tree>=1.1.3,<2.0.0',
 'tqdm>=4.64.0,<5.0.0',
 'typing_extensions>=4.0.0,<5.0.0',
 'ujson>=5.5.0,<6.0.0']

entry_points = \
{'console_scripts': ['grams = grams.cli:cli']}

setup_kwargs = {
    'name': 'sm-grams',
    'version': '2.1.7',
    'description': '',
    'long_description': '<h1 align="center">GRAMS</h1>\n\n<div align="center">\n\n![PyPI](https://img.shields.io/pypi/v/sm-grams)\n![Python](https://img.shields.io/badge/python-v3.8+-blue.svg)\n[![GitHub Issues](https://img.shields.io/github/issues/usc-isi-i2/GRAMS.svg)](https://github.com/usc-isi-i2/GRAMS/issues)\n![Contributions welcome](https://img.shields.io/badge/contributions-welcome-orange.svg)\n[![License](https://img.shields.io/badge/license-MIT-blue.svg)](https://opensource.org/licenses/MIT)\n\n</div>\n\n## Table of Contents\n- [Introduction](#introduction)\n- [Installation](#installation)\n<!-- - [Example](#example)\n- [Contributing](#contributing)\n<!-- - [Support](#support) -->\n\n## Introduction\n\nGRAMS is a library to build semantic descriptions of tables to map them to Wikidata.\n\n## Installation\n\nPlease look in the [wiki](https://github.com/usc-isi-i2/grams/wiki/Installation) for how to install GRAMS\n',
    'author': 'Binh Vu',
    'author_email': 'binh@toan2.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/usc-isi-i2/grams',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from collections import defaultdict
from typing import Dict, Optional, Callable, List, Set, Tuple, TypedDict, cast, final
from grams.algorithm.candidate_graph.cg_graph import (
    CGColumnNode,
    CGEdge,
    CGEdgeTriple,
    CGEntityValueNode,
    CGGraph,
    CGLiteralValueNode,
    CGNode,
    CGStatementNode,
)
from itertools import chain, combinations
from grams.algorithm.data_graph.dg_graph import DGGraph
from grams.algorithm.inferences.psl_lib import PSLModel
from grams.algorithm.postprocessing.common import (
    add_context,
    ensure_valid_statements,
)
from grams.algorithm.postprocessing.config import PostprocessingConfig
from grams.algorithm.postprocessing.simple_path import PostProcessingSimplePath
from grams.inputs.linked_table import LinkedTable
from operator import itemgetter
from graph.interface import EdgeTriple
from graph.retworkx.api import dag_longest_path, digraph_all_simple_paths, has_cycle
from networkx.algorithms.planarity import top_of_stack
from steiner_tree.bank.solver import BankSolver, PSEUDO_ROOT_ID
from steiner_tree.bank.struct import (
    BankEdge,
    BankGraph,
    BankNode,
    NoSingleRootException,
    Solution,
    UpwardTraversal,
)
from copy import copy


class SteinerTree:
    def __init__(
        self,
        table: LinkedTable,
        cg: CGGraph,
        dg: DGGraph,
        edge_probs: Dict[CGEdgeTriple, float],
        threshold: float,
        additional_terminal_nodes: Optional[List[str]] = None,
    ):
        self.table = table
        self.cg = cg
        self.dg = dg
        self.edge_probs = edge_probs
        self.threshold = threshold

        self.top_k_st = 50
        self.top_k_path = 50

        # extra terminal nodes that the tree should have, usually used in
        # interactive modeling where users add some entity nodes in their model
        self.additional_terminal_nodes = additional_terminal_nodes

    def get_result(self) -> CGGraph:
        """Select edges that forms a tree"""
        edge_probs = {e: p for e, p in self.edge_probs.items() if p >= self.threshold}

        # first step is to remove dangling statements and standalone nodes
        subcg = self.cg.subgraph_from_edge_triples(edge_probs.keys())
        subcg.remove_dangling_statement()
        subcg.remove_standalone_nodes()

        # terminal nodes are columns node, entity nodes are adding later
        terminal_nodes = {
            u.id for u in subcg.iter_nodes() if isinstance(u, CGColumnNode)
        }
        if self.additional_terminal_nodes is not None:
            terminal_nodes.update(self.additional_terminal_nodes)
        if len(terminal_nodes) == 0:
            # if the tree does not have any column, we essentially don't predict anything
            # so we return an empty graph
            return CGGraph()

        # add pseudo root
        edge_weights = {e: 1.0 / p for e, p in edge_probs.items()}  # p > 0.5

        solver = BankSteinerTree(
            original_graph=subcg,
            terminal_nodes=terminal_nodes,
            top_k_st=self.top_k_st,
            top_k_path=self.top_k_path,
            weight_fn=lambda e: edge_weights[e.source, e.target, e.predicate],
            solution_cmp_fn=self.compare_solutions,
            invalid_roots={
                u.id for u in subcg.nodes() if isinstance(u, CGStatementNode)
            },
            allow_shorten_graph=False,
        )
        trees, _solutions = solver.run()

        tree = cast(CGGraph, trees[0])
        tree.remove_dangling_statement()

        if PostprocessingConfig.INCLUDE_CONTEXT:
            add_context(subcg, tree, edge_probs)

        # add back statement property if missing into to ensure a correct model
        ensure_valid_statements(subcg, tree, create_if_not_exists=False)

        # fmt: off
        # from graph.viz.graphviz import draw
        # draw(graph=tree, filename="/tmp/graphviz/st204.png", **CGGraph.graphviz_props())
        # draw(graph=tree, filename="/tmp/graphviz/g25.png", **CGGraph.graphviz_props())
        # fmt: on

        return tree

    def compare_solutions(self, a: Solution, b: Solution) -> int:
        """Comparing two solutions, -1 (smaller) means a better solution -- we are solving minimum steiner tree"""
        a_weight = a.weight / max(a.num_edges, 1)
        b_weight = b.weight / max(b.num_edges, 1)

        if a_weight < b_weight:
            return -1
        if a_weight > b_weight:
            return 1
        # equal weight, prefer the one with shorter depth
        if not hasattr(a, "depth"):
            setattr(a, "depth", len(dag_longest_path(a.graph)))
        if not hasattr(b, "depth"):
            setattr(b, "depth", len(dag_longest_path(b.graph)))
        return getattr(a, "depth") - getattr(b, "depth")


class BankSteinerTree(BankSolver[CGNode, CGEdge]):
    ADD_MISSING_STATEMENT_PROPS = True

    def _solve(
        self, g: BankGraph, terminal_nodes: Set[str], top_k_st: int, top_k_path: int
    ):
        """Override the main algorithm to handle statement"""
        roots = {u.id for u in g.iter_nodes() if u.id not in self.invalid_roots}

        attr_visit_hists: List[Tuple[str, UpwardTraversal]] = []
        # to ensure the order
        for uid in list(sorted(terminal_nodes)):
            visit_hist = UpwardTraversal.top_k_beamsearch(g, uid, top_k_path)
            roots = roots.intersection(visit_hist.paths.keys())
            attr_visit_hists.append((uid, visit_hist))

        if len(roots) == 0:
            raise NoSingleRootException()

        # to ensure the order again & remove randomness
        roots = sorted(roots)

        # merge the paths using beam search
        results = []
        for root in roots:
            current_states = []
            uid, visit_hist = attr_visit_hists[0]
            for path in visit_hist.paths[root]:
                pg = BankGraph()
                if len(path.path) > 0:
                    assert uid == path.path[0].target
                pg.add_node(BankNode(uid))
                for e in path.path:
                    pg.add_node(BankNode(e.source))
                    pg.add_edge(e.clone())

                self.add_missing_statement(pg)
                current_states.append(pg)

            if len(current_states) > top_k_st:
                current_states = [
                    _s.graph for _s in self._sort_solutions(current_states)[:top_k_st]
                ]

            for uid, visit_hist in attr_visit_hists[1:]:
                next_states = []
                for state in current_states:
                    for path in visit_hist.paths[root]:
                        pg = state.copy()
                        if len(path.path) > 0:
                            assert uid == path.path[0].target
                        if not pg.has_node(uid):
                            pg.add_node(BankNode(uid))
                        for e in path.path:
                            if not pg.has_node(e.source):
                                pg.add_node(BankNode(id=e.source))
                            # TODO: here we don't check by edge_key because we may create another edge of different key
                            # hope this new path has been exploited before.
                            if not pg.has_edges_between_nodes(e.source, e.target):
                                pg.add_edge(e.clone())
                        self.add_missing_statement(pg)

                        # if there are more than path between two nodes within
                        # two hop, we'll select one
                        update_graph = False
                        for n in pg.iter_nodes():
                            if pg.in_degree(n.id) >= 2:
                                grand_parents: Dict[
                                    str, List[Tuple[BankEdge, ...]]
                                ] = defaultdict(list)
                                for inedge in pg.in_edges(n.id):
                                    grand_parents[inedge.source].append((inedge,))
                                    for grand_inedge in pg.in_edges(inedge.source):
                                        grand_parents[grand_inedge.source].append(
                                            (grand_inedge, inedge)
                                        )

                                for grand_parent, edges in grand_parents.items():
                                    if len(edges) > 1:
                                        # we need to select one path from this grand parent to the rest
                                        # they have the same length, so we select the one has smaller weight
                                        edges = sorted(
                                            edges,
                                            key=lambda x: x[0].weight + x[1].weight
                                            if len(x) == 2
                                            else x[0].weight * 2,
                                        )

                                        for lst in edges[1:]:
                                            for edge in lst:
                                                # TODO: handle removing edges multiple times
                                                try:
                                                    pg.remove_edge(edge.id)
                                                except:
                                                    continue
                                        update_graph = True
                        if update_graph:
                            for n in pg.nodes():
                                if pg.in_degree(n.id) == 0 and pg.out_degree(n.id) == 0:
                                    pg.remove_node(n.id)
                        # after add a path to the graph, it can create new cycle, detect and fix it
                        if has_cycle(pg):
                            # we can show that if the graph contain cycle, there is a better path
                            # so no need to try to break cycles as below
                            # cycles_iter = [(uid, vid) for uid, vid, eid, orien in cycles_iter]
                            # for _g in self._break_cycles(root, pg, cycles_iter):
                            #     next_states.append(_g)
                            pass
                        else:
                            next_states.append(pg)

                        # the output graph should not have parallel edges
                        assert not pg.has_parallel_edges()
                if len(next_states) > top_k_st:
                    next_states = [
                        _s.graph for _s in self._sort_solutions(next_states)[:top_k_st]
                    ]
                # assert all(_.check_integrity() for _ in next_states)
                current_states = next_states
                # cgs = [g for g in next_states if len(list(nx.simple_cycles(g))) > 0]
                # nx.draw_networkx(cgs[0]); plt.show()
                # nx.draw(cgs[0]); plt.show()
            results += current_states

        return self._sort_solutions(results)

    def add_missing_statement(self, g: BankGraph):
        if not self.ADD_MISSING_STATEMENT_PROPS:
            return

        cg = cast(CGGraph, self.original_graph)
        for uprime in g.iter_nodes():
            if uprime.id == PSEUDO_ROOT_ID:
                continue
            u = cg.get_node(uprime.id)
            if isinstance(u, CGStatementNode):
                (inedge,) = cg.in_edges(u.id)
                if all(outedge.key != inedge.key for outedge in g.out_edges(u.id)):
                    # no statement property, add it back
                    (main_prop,) = [
                        outedge
                        for outedge in self.graph.out_edges(u.id)
                        if outedge.key == inedge.key
                    ]
                    if not g.has_node(main_prop.target):
                        target_id = g.add_node(
                            copy(self.graph.get_node(main_prop.target))
                        )
                        assert target_id == main_prop.target
                    g.add_edge(main_prop.clone())



from collections import defaultdict
from dataclasses import dataclass
from typing import (
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Tuple,
    TypeVar,
)
from operator import attrgetter
from grams.algorithm.data_graph.dg_graph import DGGraph, DGNode, EntityValueNode
from grams.algorithm.inferences.psl_lib import IDMap

from grams.algorithm.data_graph import CellNode
from grams.algorithm.literal_matchers import TextParser
from grams.algorithm.candidate_graph.cg_graph import (
    CGColumnNode,
    CGEdge,
    CGEntityValueNode,
    CGGraph,
    CGLiteralValueNode,
    CGNode,
    CGStatementNode,
)
from grams.inputs.linked_table import LinkedTable
from kgdata.wikidata.models import (
    WDEntity,
    WDProperty,
    WDQuantityPropertyStats,
    WDEntityLabel,
    WDClass,
)
from sm.misc.fn_cache import CacheMethod


K = TypeVar("K")
V = TypeVar("V")


class RelFeatures:
    def __init__(
        self,
        idmap: IDMap,
        table: LinkedTable,
        cg: CGGraph,
        dg: DGGraph,
        wdentities: Mapping[str, WDEntity],
        wdentity_labels: Mapping[str, WDEntityLabel],
        wdclasses: Mapping[str, WDClass],
        wdprops: Mapping[str, WDProperty],
        wd_num_prop_stats: Mapping[str, WDQuantityPropertyStats],
        sim_fn: Optional[Callable[[str, str], float]] = None,
    ):
        self.idmap = idmap
        self.table = table
        self.cg = cg
        self.dg = dg
        self.wdentities = wdentities
        self.wdclasses = wdclasses
        self.wdprops = wdprops
        self.wd_num_prop_stats = wd_num_prop_stats
        self.sim_fn = sim_fn

    def extract_features(self, features: List[str]) -> Dict[str, list]:
        rels = []
        for s in self.cg.iter_nodes():
            if not isinstance(s, CGStatementNode):
                continue
            (inedge,) = self.cg.in_edges(s.id)
            for outedge in self.cg.out_edges(s.id):
                rels.append((s, inedge, outedge))

        feat_data = {}
        for feat in features:
            fn = getattr(self, feat)
            feat_data[feat] = fn(rels)
        return feat_data

    def REL_FREQ_OVER_ROW(self, rels: List[Tuple[CGStatementNode, CGEdge, CGEdge]]):
        """The frequency of the relation over the row."""
        n_rows = self.table.size()
        output = []

        for s, inedge, outedge in rels:
            ratio = self.get_rel_freq(s, outedge) / n_rows
            if inedge.predicate == outedge.predicate:
                output.append(
                    (
                        self.idmap.m(inedge.source),
                        self.idmap.m(s.id),
                        self.idmap.m(inedge.predicate),
                        ratio,
                    )
                )
            output.append(
                (
                    self.idmap.m(s.id),
                    self.idmap.m(outedge.target),
                    self.idmap.m(outedge.predicate),
                    ratio,
                )
            )
        return output

    def REL_FREQ_OVER_ENT_ROW(self, rels: List[Tuple[CGStatementNode, CGEdge, CGEdge]]):
        # what is the maximum possible links we can have? this ignore the the link so this is used to calculate FreqOverEntRow
        output = []
        for s, inedge, outedge in rels:
            max_pos_ent_rows = self.get_maximum_possible_ent_links_between_two_nodes(
                s, inedge, outedge
            )
            freq = self.get_rel_freq(s, outedge)
            ratio = 0 if max_pos_ent_rows == 0 else freq / max_pos_ent_rows
            if inedge.predicate == outedge.predicate:
                output.append(
                    (
                        self.idmap.m(inedge.source),
                        self.idmap.m(s.id),
                        self.idmap.m(inedge.predicate),
                        ratio,
                    )
                )
            output.append(
                (
                    self.idmap.m(s.id),
                    self.idmap.m(outedge.target),
                    self.idmap.m(outedge.predicate),
                    ratio,
                )
            )
        return output

    def REL_FREQ_OVER_POS_REL(self, rels: List[Tuple[CGStatementNode, CGEdge, CGEdge]]):
        """The frequency of the relation over all possible relations that the two nodes can have."""
        output = []
        tmp = {}
        max_possible_links = {}

        for s, inedge, outedge in rels:
            freq = self.get_rel_freq(s, outedge)
            n_unmatch_links = self.get_unmatch_discovered_links(s, inedge, outedge)
            n_possible_links = n_unmatch_links + len(self.get_rel_dg_pairs(s, outedge))

            max_possible_links[inedge.source, outedge.target] = max(
                n_possible_links,
                max_possible_links.get((inedge.source, outedge.target), 0),
            )

            tmp[
                (
                    self.idmap.m(s.id),
                    self.idmap.m(outedge.target),
                    self.idmap.m(outedge.predicate),
                )
            ] = (freq, (inedge.source, outedge.target))
            if inedge.predicate == outedge.predicate:
                tmp[
                    (
                        self.idmap.m(inedge.source),
                        self.idmap.m(s.id),
                        self.idmap.m(inedge.predicate),
                    )
                ] = (freq, (inedge.source, outedge.target))

        for key, (freq, pair) in tmp.items():
            prob = (
                0.0
                if max_possible_links[pair] == 0
                else freq / max_possible_links[pair]
            )
            output.append(key + (prob,))

        return output

    def REL_FREQ_UNMATCH_OVER_ENT_ROW(
        self, rels: List[Tuple[CGStatementNode, CGEdge, CGEdge]]
    ):
        output = []
        for s, inedge, outedge in rels:
            max_pos_ent_rows = self.get_maximum_possible_ent_links_between_two_nodes(
                s, inedge, outedge
            )
            n_unmatch_links = self.get_unmatch_discovered_links(s, inedge, outedge)
            ratio = 0 if max_pos_ent_rows == 0 else n_unmatch_links / max_pos_ent_rows
            if inedge.predicate == outedge.predicate:
                output.append(
                    (
                        self.idmap.m(inedge.source),
                        self.idmap.m(s.id),
                        self.idmap.m(inedge.predicate),
                        ratio,
                    )
                )
            output.append(
                (
                    self.idmap.m(s.id),
                    self.idmap.m(outedge.target),
                    self.idmap.m(outedge.predicate),
                    ratio,
                )
            )
        return output

    def REL_FREQ_UNMATCH_OVER_POS_REL(
        self, rels: List[Tuple[CGStatementNode, CGEdge, CGEdge]]
    ):
        output = []
        tmp = {}
        max_possible_links = {}

        for s, inedge, outedge in rels:
            n_unmatch_links = self.get_unmatch_discovered_links(s, inedge, outedge)
            n_possible_links = n_unmatch_links + len(self.get_rel_dg_pairs(s, outedge))

            max_possible_links[inedge.source, outedge.target] = max(
                n_possible_links,
                max_possible_links.get((inedge.source, outedge.target), 0),
            )

            tmp[
                (
                    self.idmap.m(s.id),
                    self.idmap.m(outedge.target),
                    self.idmap.m(outedge.predicate),
                )
            ] = (n_unmatch_links, (inedge.source, outedge.target))
            if inedge.predicate == outedge.predicate:
                tmp[
                    (
                        self.idmap.m(inedge.source),
                        self.idmap.m(s.id),
                        self.idmap.m(inedge.predicate),
                    )
                ] = (n_unmatch_links, (inedge.source, outedge.target))

        for key, (freq, pair) in tmp.items():
            prob = (
                0.0
                if max_possible_links[pair] == 0
                else freq / max_possible_links[pair]
            )
            output.append(key + (prob,))

        return output

    def REL_HEADER_SIMILARITY(self, rels: List[Tuple[CGStatementNode, CGEdge, CGEdge]]):
        if self.sim_fn is None:
            return []

        props = self.get_props()
        output = []
        for s, inedge, outedge in rels:
            target = self.cg.get_node(outedge.target)
            if not isinstance(target, CGColumnNode):
                continue
            sim = self.sim_fn(target.label, props[outedge.predicate].label)

            if inedge.predicate == outedge.predicate:
                output.append(
                    (
                        self.idmap.m(inedge.source),
                        self.idmap.m(inedge.target),
                        self.idmap.m(inedge.predicate),
                        sim,
                    )
                )
            output.append(
                (
                    self.idmap.m(outedge.source),
                    self.idmap.m(outedge.target),
                    self.idmap.m(outedge.predicate),
                    sim,
                )
            )
        return output

    def REL_NOT_FUNC_DEPENDENCY(
        self, rels: List[Tuple[CGStatementNode, CGEdge, CGEdge]]
    ):
        output = []
        for s, inedge, outedge in rels:
            u = self.cg.get_node(inedge.source)
            v = self.cg.get_node(outedge.target)
            if not isinstance(u, CGColumnNode) or not isinstance(v, CGColumnNode):
                continue

            not_func_dep = int(not self.is_functional_dependency(u.column, v.column))
            if inedge.predicate == outedge.predicate:
                output.append(
                    (
                        self.idmap.m(inedge.source),
                        self.idmap.m(s.id),
                        self.idmap.m(inedge.predicate),
                        not_func_dep,
                    )
                )
            output.append(
                (
                    self.idmap.m(s.id),
                    self.idmap.m(outedge.target),
                    self.idmap.m(outedge.predicate),
                    not_func_dep,
                )
            )
        return output

    @CacheMethod.cache(CacheMethod.two_object_args)
    def get_rel_freq(self, s: CGStatementNode, outedge: CGEdge):
        """Get frequency of the link, which is sum of links all rows between two nodes"""
        dg_flows = s.get_edges_provenance([outedge])
        sum_prob = 0.0
        for source_flow, target_flows in dg_flows.items():
            # regardless of entity or column node, we count all links between rows
            # for column node, we only have one for each row (one target_flows),
            # for entity node, we only have one source flow, but multiple target flows (max one per row)
            for target_flow, provenances in target_flows.items():
                sum_prob += max(provenances, key=attrgetter("prob")).prob
        return sum_prob

    @CacheMethod.cache(CacheMethod.two_object_args)
    def get_rel_dg_pairs(self, s: CGStatementNode, outedge: CGEdge):
        return {
            (source_flow.dg_source_id, target_flow.dg_target_id)
            for source_flow, target_flows in s.get_edges_provenance([outedge]).items()
            for target_flow, provenances in target_flows.items()
        }

    @CacheMethod.cache(CacheMethod.three_object_args)
    def get_maximum_possible_ent_links_between_two_nodes(
        self, s: CGStatementNode, inedge: CGEdge, outedge: CGEdge
    ):
        """Find the maximum possible links between two nodes (ignore the possible predicates):

        Let M be the maximum possible links we want to find, N is the number of rows in the table.
        1. If two nodes are not columns, M is 1 because it's entity to entity link.
        2. If one node is a column, M = N - U, where U is the number of pairs that cannot have KG discovered links. A
        pair that cannot have KG discovered links is:
            a. If both nodes are columns, and the link is
                * data predicate: the source cell links to no entity.
                * object predicate: the source or target cell link to no entity
            b. If only one node is column, and the link is
                * data predicate:
                    - if the source node must be an entity, then the target must be a column. U is always 0
                    - else then the source node must be is a column and target is a literal value: a cell in the column links to no entity
                * object predicate: a cell in the column links to no entity.
        """
        u = self.cg.get_node(inedge.source)
        v = self.cg.get_node(outedge.target)

        if not isinstance(u, CGColumnNode) and not isinstance(v, CGColumnNode):
            return 1

        # instead of going through each node attach to the node in the semantic graph, we avoid by directly generating
        # the data node ID
        n_rows = self.table.size()
        n_null_entities = 0
        is_data_predicate = self.wdprops[outedge.predicate].is_data_property()

        if not (isinstance(u, CGColumnNode) and isinstance(v, CGColumnNode)):
            if is_data_predicate:
                if isinstance(u, (CGEntityValueNode, CGLiteralValueNode)):
                    assert isinstance(u, CGEntityValueNode) and isinstance(
                        v, CGColumnNode
                    )
                    return n_rows

                assert isinstance(u, CGColumnNode) and isinstance(
                    v, (CGEntityValueNode, CGLiteralValueNode)
                )

            if isinstance(u, CGColumnNode):
                ci = u.column
            else:
                assert isinstance(
                    v, CGColumnNode
                ), "Always true -- type system is not smart enough"
                ci = v.column

            for ri in range(n_rows):
                if len(self.dg.get_cell_node(f"{ri}-{ci}").entity_ids) == 0:
                    n_null_entities += 1
        else:
            uci = u.column
            vci = v.column

            for ri in range(n_rows):
                ucell_unk = len(self.dg.get_cell_node(f"{ri}-{uci}").entity_ids) == 0
                vcell_unk = len(self.dg.get_cell_node(f"{ri}-{vci}").entity_ids) == 0

                if is_data_predicate:
                    if ucell_unk:
                        n_null_entities += 1
                elif ucell_unk or vcell_unk:
                    n_null_entities += 1

        return n_rows - n_null_entities

    @CacheMethod.cache(CacheMethod.three_object_args)
    def get_unmatch_discovered_links(
        self, s: CGStatementNode, inedge: CGEdge, outedge: CGEdge
    ):
        """Get number of discovered links that don't match due to value differences. This function do not count if:
        * the link between two DG nodes is impossible
        * the property/qualifier do not exist in the WDEntity
        """
        u = self.cg.get_node(inedge.source)
        v = self.cg.get_node(outedge.target)
        uv_links = self.get_rel_dg_pairs(s, outedge)
        is_outpred_qualifier = inedge.predicate != outedge.predicate
        is_outpred_data_predicate = self.wdprops[outedge.predicate].is_data_property()

        n_unmatch_links = 0
        for dgu, dgv in self._iter_dg_pair(inedge.source, outedge.target):
            # if has link, then we don't have to count
            if (dgu.id, dgv.id) in uv_links:
                continue

            # ignore pairs that can't have any links
            if not self._dg_pair_has_possible_ent_links(
                dgu, dgv, is_outpred_data_predicate
            ):
                continue

            if isinstance(dgu, CellNode):
                # the source is cell node
                # property doesn't exist in any qnode
                dgu_qnodes = [self.wdentities[qnode_id] for qnode_id in dgu.entity_ids]
                if all(inedge.predicate not in qnode.props for qnode in dgu_qnodes):
                    continue

                if is_outpred_qualifier:
                    # qualifier doesn't exist in any qnode
                    has_qual = False
                    for qnode in dgu_qnodes:
                        for stmt in qnode.props.get(inedge.predicate, []):
                            if outedge.predicate in stmt.qualifiers:
                                has_qual = True
                    if not has_qual:
                        continue

                # # #########################
                # # TODO: remove me, modify on Apr 29, should have a better solution
                # # this is to say like if that properties have multiple values, then it's more likely to miss
                # if all(len(qnode.props.get(inpred, [])) > 2 for qnode in dgu_qnodes):
                #     continue
                # # #########################
            else:
                # the source is entity
                assert isinstance(dgu, EntityValueNode)
                dgu_qnode = self.wdentities[dgu.qnode_id]
                if inedge.predicate not in dgu_qnode.props:
                    continue
                if is_outpred_qualifier:
                    if all(
                        outedge.predicate not in stmt.qualifiers
                        for stmt in dgu_qnode.props[inedge.predicate]
                    ):
                        continue
                # # #########################
                # # TODO: remove me, modify on Apr 29, should have a better solution
                # # this is to say like if that properties have multiple values, then it's more likely to miss
                # if len(dgu_qnode.props.get(inpred, [])) > 2:
                #     continue
                # # #########################
            n_unmatch_links += 1
        return n_unmatch_links

    @CacheMethod.cache(CacheMethod.as_is_posargs)
    def is_functional_dependency(
        self, source_column_index: int, target_column_index: int
    ) -> bool:
        """Test whether values in the target column is uniquely determiend by the values in the source column. True if
        it's FD.

        Parameters
        ----------
        source_column_index
        target_column_index

        Returns
        -------
        """
        sci = source_column_index
        tci = target_column_index

        # find a mapping from
        source_map = self._get_value_map(source_column_index)
        target_map = {
            ri: key
            for key, rows in self._get_value_map(target_column_index).items()
            for ri in rows
        }

        n_violate_fd = 0
        for key, rows in source_map.items():
            target_keys = {target_map[ri] for ri in rows}
            if len(target_keys) > 1:
                n_violate_fd += 1

        if len(source_map) == 0:
            return True

        if n_violate_fd / len(source_map) > 0.01:
            return False
        return True

    def _iter_dg_pair(self, uid: str, vid: str) -> Iterable[Tuple[DGNode, DGNode]]:
        """This function iterate through each pair of data graph nodes between two semantic graph nodes.

        If both sg nodes are entities, we only have one pair.
        If one or all of them are columns, the number of pairs will be the size of the table.
        Otherwise, not support iterating between nodes & statements
        """
        u = self.cg.get_node(uid)
        v = self.cg.get_node(vid)

        if isinstance(u, CGColumnNode) and isinstance(v, CGColumnNode):
            uci = u.column
            vci = v.column
            for ri in range(self.table.size()):
                ucell = self.dg.get_node(f"{ri}-{uci}")
                vcell = self.dg.get_node(f"{ri}-{vci}")
                yield ucell, vcell
        elif isinstance(u, CGColumnNode):
            assert isinstance(v, (CGEntityValueNode, CGLiteralValueNode))
            uci = u.column
            vcell = self.dg.get_node(v.id)
            for ri in range(self.table.size()):
                ucell = self.dg.get_node(f"{ri}-{uci}")
                yield ucell, vcell
        elif isinstance(v, CGColumnNode):
            assert isinstance(u, (CGEntityValueNode, CGLiteralValueNode))
            vci = v.column
            ucell = self.dg.get_node(u.id)
            for ri in range(self.table.size()):
                vcell = self.dg.get_node(f"{ri}-{vci}")
                yield ucell, vcell
        else:
            assert not isinstance(u, CGColumnNode) and not isinstance(v, CGColumnNode)
            yield self.dg.get_node(u.id), self.dg.get_node(v.id)

    def _dg_pair_has_possible_ent_links(
        self, dgu: DGNode, dgv: DGNode, is_data_predicate: bool
    ):
        if isinstance(dgu, CellNode) and isinstance(dgv, CellNode):
            # both are cells
            if is_data_predicate:
                # data predicate: source cell must link to some entities to have possible links
                return len(dgu.entity_ids) > 0
            else:
                # object predicate: source cell and target cell must link to some entities to have possible links
                return len(dgu.entity_ids) > 0 and len(dgv.entity_ids) > 0
        elif isinstance(dgu, CellNode):
            # the source is cell, the target will be literal/entity value
            # we have link when source cell link to some entities, doesn't depend on type of predicate
            return len(dgu.entity_ids) > 0
        elif isinstance(dgv, CellNode):
            # the target is cell, the source will be literal/entity value
            if is_data_predicate:
                # data predicate: always has possibe links
                return True
            else:
                # object predicate: have link when the target cell link to some entities
                return len(dgv.entity_ids) > 0
        else:
            # all cells are values, always have link due to how the link is generated in the first place
            return True

    @CacheMethod.cache(CacheMethod.as_is_posargs)
    def _get_value_map(self, column_index: int):
        """Get a map of values in a column to its row numbers (possible duplication).
        This function is not perfect now. The value of the column is considered to be list of entities (if exist) or
        just the value of the cell
        """
        map = defaultdict(list)
        col = self.table.table.get_column_by_index(column_index)
        for ri in range(self.table.size()):
            links = self.table.links[ri][column_index]
            ents = [link.entity_id for link in links if link.entity_id is not None]
            if len(ents) > 0:
                key = tuple(ents)
            else:
                key = col.values[ri].strip()
            map[key].append(ri)
        return dict(map)

    @CacheMethod.cache(CacheMethod.as_is_posargs)
    def get_props(self) -> Dict[str, WDProperty]:
        """Get properties used in the candidate graph"""
        prop_ids = {edge.predicate for edge in self.cg.iter_edges()}
        return {prop_id: self.wdprops[prop_id] for prop_id in prop_ids}



from __future__ import annotations
from typing import Any, Callable, Dict, Iterable, List, Literal, Optional, Set, Union
from dataclasses import dataclass
from copy import copy
from typing_extensions import Self

from grams.algorithm.inferences.features.type_feature_config import TypeFeatureConfigs


@dataclass
class TreeStruct:
    __slots__ = ("p2cs", "c2ps")
    p2cs: Dict[str, Set[str]]  # from parent to children
    c2ps: Dict[str, Set[str]]  # from child to parents

    @staticmethod
    def construct(
        items: Union[Set[str], Dict[str, Any]],
        get_parents: Callable[[str], Iterable[str]],
    ):
        p2cs = {x: set() for x in items}
        c2ps = {x: set() for x in items}

        for item in items:
            parents = get_parents(item)
            for p in parents:
                if p in items:
                    p2cs[p].add(item)
                    c2ps[item].add(p)

        return TreeStruct(p2cs, c2ps)

    def ensure_tree(
        self, fix_method: Optional[Literal["auto", "manually"]] = None
    ) -> Self:
        """Make sure this is actually a poly-tree, not containing any cycle
        If we found a cycle, we remove one link to make it acyclic.
        """
        visited = set()
        cycles = []

        for c, ps in self.c2ps.items():
            if len(ps) == 0:
                path = {}
                backref = self._dfs_find_backref(c, path, visited)
                if backref is None:
                    continue
                cycles.append(self._get_cycles(path, backref))

        if len(visited) == len(self.c2ps) and len(cycles) == 0:
            # all nodes have been visited, and no cycles found
            return self

        if fix_method == "manually" or (
            fix_method is None and TypeFeatureConfigs.FIX_TYPE_CYCLE == "manually"
        ):
            if len(cycles) == 0:
                for c in self.c2ps:
                    if c in visited:
                        continue
                    path = {}
                    backref = self._dfs_find_backref(c, path, visited)
                    if backref is None:
                        continue
                    cycles.append(self._get_cycles(path, backref))
            raise ValueError(f"Found cycles: {cycles}")

        if len(cycles) == 0:
            for c in self.c2ps:
                if c in visited:
                    continue
                path = {}
                backref = self._dfs_find_backref(c, path, visited)
                if backref is None:
                    continue
                cycles.append(self._get_cycles(path, backref))

        if fix_method == "manually" or (
            fix_method is None and TypeFeatureConfigs.FIX_TYPE_CYCLE == "manually"
        ):
            raise ValueError(f"Found cycles: {cycles}")

        # those cycles are disjoint, so we can break them one by one
        for cycle in cycles:
            self._break_cycles(cycle)

        return self.ensure_tree()

    def _dfs_find_backref(
        self, node: str, path: Dict[str, int], visited: Set[str]
    ) -> Optional[str]:
        """Using DFS to find back reference edge in the tree"""
        if node in path:
            return node

        # visiting this node
        path[node] = len(path)
        visited.add(node)

        for child in self.p2cs[node]:
            if (resp := self._dfs_find_backref(child, path, visited)) is not None:
                return resp

        path.pop(node)
        return None

    def _get_cycles(self, path: Dict[str, int], backref: str) -> List[str]:
        """Extract cycles from DFS results."""
        items = list(path.items())
        # just want to make sure that the items are in order
        assert all(items[i][1] + 1 == items[i + 1][1] for i in range(len(items) - 1))

        idx = next(i for i in range(len(items)) if items[i][0] == backref)
        # remove the previous edges that does not create cycles
        return [x[0] for x in items[idx:]]

    def _break_cycles(self, cycle: List[str]):
        # we find the node that has the most children to remove
        best_item_index = max(
            range(len(cycle)), key=lambda x: len(self.p2cs[cycle[x][0]])
        )
        source = cycle[best_item_index][0]
        target = cycle[(best_item_index + 1) % len(cycle)][0]
        self.p2cs[source].remove(target)
        self.c2ps[target].remove(source)
